# -*- coding: utf-8 -*-

import pytest
from .as_status_codes import AerospikeStatus
from .index_helpers import ensure_dropped_index
from .test_base_class import TestBaseClass
from aerospike import exception as e

import aerospike


class TestIndex(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "addr": "name%s" % (str(i)), "age": i, "no": i}
            as_connection.put(key, rec)

        def teardown():
            ensure_dropped_index(self.as_connection, "test", "age_index")
            ensure_dropped_index(self.as_connection, "test", "name_index")
            for i in range(5):
                key = ("test", "demo", i)
                # TODO: unneeded variable?
                rec = {"name": "name%s" % (str(i)), "addr": "name%s" % (str(i)), "age": i, "no": i}  # noqa: F841
                as_connection.remove(key)

        request.addfinalizer(teardown)

    def test_create_indexes_with_no_parameters(self):
        """
        Invoke indexc_string_reate() without any
        mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.index_string_create()

        assert "argument 'ns' (pos 1)" in str(typeError.value)

        with pytest.raises(TypeError) as typeError:
            self.as_connection.index_integer_create()

        assert "argument 'ns' (pos 1)" in str(typeError.value)

    def test_create_integer_index_with_correct_parameters(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_integer_index_with_set_name_too_long(self):
        # Invoke createindex with a set name beyond the maximum
        set_name = "a" * 128

        policy = {}
        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_integer_create("test", set_name, "age", "age_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_create_integer_index_with_incorrect_namespace(self):
        """
        Invoke createindex() with non existent namespace
        """
        policy = {}
        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_integer_create("fake_namespace", "demo", "age", "age_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_create_integer_index_with_incorrect_set(self):
        """
        Invoke createindex() with nonexistent set
        It should succeed
        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo1", "age", "age_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_integer_index_with_incorrect_bin(self):
        """
        Invoke createindex() with a nonexistent bin
        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo", "fake_bin", "age_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_integer_index_with_namespace_is_none(self):
        """
        Invoke createindex() with namespace is None
        """
        policy = {}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_integer_create(None, "demo", "age", "age_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_creat_integer_eindex_with_set_is_none(self):
        # Invoke createindex() with set is None
        policy = {}

        retobj = self.as_connection.index_integer_create("test", None, "age", "age_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_integer_index_with_set_is_int(self):
        # Invoke createindex() with set is int
        policy = {}

        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_integer_create("test", 1, "age", "age_index", policy)
        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_integer_index_with_bin_is_none(self):
        """
        Invoke createindex() with bin is None
        """
        policy = {}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_integer_create("test", "demo", None, "age_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_integer_index_with_index_is_none(self):
        """
        Invoke createindex() with index_name is None
        """
        policy = {}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_integer_create("test", "demo", "age", None, policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_same_integer_index_multiple_times(self):
        """
        Invoke createindex() with the same arguments
        multiple times on the same bin

        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)
        assert retobj == AerospikeStatus.AEROSPIKE_OK
        try:
            retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)
        except e.IndexFoundError:
            assert self.server_version <= [6, 0]

        self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_same_integer_index_multiple_times_different_bin(self):
        """
        Invoke createindex() with the same index name,
        multiple times on different bin names
        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK

        with pytest.raises(e.IndexFoundError):
            retobj = self.as_connection.index_integer_create("test", "demo", "no", "age_index", policy)
            self.as_connection.index_remove("test", "age_index", policy)

        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_different_integer_index_multiple_times_same_bin(self):
        """
                    Invoke createindex() with multiple times on same bin with different
        name
        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)
        assert retobj == AerospikeStatus.AEROSPIKE_OK
        try:
            retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index1", policy)
            self.as_connection.index_remove("test", "age_index1", policy)
        except e.IndexFoundError:
            assert self.server_version <= [6, 0]

        ensure_dropped_index(self.as_connection, "test", "age_index")

    def test_create_integer_index_with_policy(self):
        """
        Invoke createindex() with policy
        """
        policy = {"timeout": 180000}
        retobj = self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)

        ensure_dropped_index(self.as_connection, "test", "age_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_create_string_index_positive(self):
        """
        Invoke create string index() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index", policy)

        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")

        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_create_string_index_with_set_length_too_long(self):
        # Invoke createindex() with correct arguments set length extra
        set_name = "a" * 100
        policy = {}

        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_string_create("test", set_name, "name", "name_index", policy)
        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_create_string_index_with_correct_parameters_ns_length_extra(self):
        # Invoke createindex() with correct arguments ns length extra
        ns_name = "a" * 50
        policy = {}

        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_string_create(ns_name, "demo", "name", "name_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_create_string_index_with_incorrect_namespace(self):
        """
        Invoke create string index() with incorrect namespace
        """
        policy = {}

        with pytest.raises(e.InvalidRequest) as err_info:
            self.as_connection.index_string_create("fake_namespace", "demo", "name", "name_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_REQUEST_INVALID

    def test_create_string_index_with_incorrect_set(self):
        """
        Invoke create string index() with incorrect set
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo1", "name", "name_index", policy)

        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_create_string_index_with_incorrect_bin(self):
        """
        Invoke create string index() with incorrect bin
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo", "name1", "name_index", policy)

        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_create_string_index_with_namespace_is_none(self):
        """
        Invoke create string index() with namespace is None
        """
        policy = {}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_string_create(None, "demo", "name", "name_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_string_index_with_set_is_none(self):
        # Invoke create string index() with set is None
        policy = {}
        retobj = self.as_connection.index_string_create("test", None, "name", "name_index", policy)

        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_create_string_index_with_bin_is_none(self):
        """
        Invoke create string index() with bin is None
        """
        policy = {}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_string_create("test", "demo", None, "name_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_string_index_with_index_is_none(self):
        """
        Invoke create_string_index() with index name is None
        """
        policy = {}
        with pytest.raises(e.ParamError) as err_info:
            self.as_connection.index_string_create("test", "demo", "name", None, policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_ERR_PARAM

    def test_create_same_string_index_multiple_times(self):
        """
        Invoke create string index() with multiple times on same bin
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index", policy)
        assert retobj == AerospikeStatus.AEROSPIKE_OK
        try:
            retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index", policy)
        except e.IndexFoundError:
            assert self.server_version <= [6, 0]

        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")

    def test_create_same_string_index_multiple_times_different_bin(self):
        """
        Invoke create string index() with multiple times on different bin
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index", policy)
        assert retobj == AerospikeStatus.AEROSPIKE_OK

        with pytest.raises(e.IndexFoundError):
            retobj = self.as_connection.index_string_create("test", "demo", "addr", "name_index", policy)
            self.as_connection.index_remove("test", "name_index", policy)
            ensure_dropped_index(self.as_connection, "test", "name_index")

        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")

        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_create_different_string_index_multiple_times_same_bin(self):
        """
        Invoke create string index() with multiple times on same
        bin with different name
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index", policy)
        assert retobj == AerospikeStatus.AEROSPIKE_OK
        try:
            retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index1", policy)
            self.as_connection.index_remove("test", "name_index1", policy)
        except e.IndexFoundError:
            assert self.server_version <= [6, 0]

        ensure_dropped_index(self.as_connection, "test", "name_index")

    def test_create_string_index_with_policy(self):
        """
        Invoke create string index() with policy
        """
        policy = {"timeout": 180000}
        retobj = self.as_connection.index_string_create("test", "demo", "name", "name_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "name_index")

    def test_drop_invalid_index(self):
        """
        Invoke drop invalid index()
        """
        policy = {}
        try:
            self.as_connection.index_remove("test", "notarealindex", policy)
        except e.IndexNotFound:
            assert self.server_version <= [6, 0]

    def test_drop_valid_index(self):
        """
        Invoke drop valid index()
        """
        policy = {}
        self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)
        retobj = self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_drop_valid_index_policy(self):
        """
        Invoke drop valid index() policy
        """
        policy = {"timeout": 180000}
        self.as_connection.index_integer_create("test", "demo", "age", "age_index", policy)
        retobj = self.as_connection.index_remove("test", "age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "age_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_createindex_with_long_index_name(self):
        # Invoke createindex() with long index name
        policy = {}
        with pytest.raises(e.InvalidRequest):
            self.as_connection.index_integer_create("test", "demo", "age", "index" * 100, policy)

    def test_create_string_index_unicode_positive(self):
        """
        Invoke create string index() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_string_create("test", "demo", "name", "uni_name_index", policy)

        self.as_connection.index_remove("test", "uni_name_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_name_index")
        assert retobj == AerospikeStatus.AEROSPIKE_OK

    def test_createindex_integer_unicode(self):
        """
        Invoke createindex() with correct arguments
        """
        policy = {}
        retobj = self.as_connection.index_integer_create("test", "demo", "age", "uni_age_index", policy)

        assert retobj == AerospikeStatus.AEROSPIKE_OK
        self.as_connection.index_remove("test", "uni_age_index", policy)
        ensure_dropped_index(self.as_connection, "test", "uni_age_index")

    def test_createindex_with_correct_parameters_without_connection(self):
        # Invoke createindex() with correct arguments without connection
        policy = {}
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        with pytest.raises(e.ClusterError) as err_info:
            client1.index_integer_create("test", "demo", "age", "age_index", policy)

        err_code = err_info.value.code
        assert err_code is AerospikeStatus.AEROSPIKE_CLUSTER_ERROR

    def test_index_remove_no_args(self):

        with pytest.raises(TypeError):
            self.as_connection.index_remove()

    def test_index_remove_no_index(self):

        with pytest.raises(TypeError):
            self.as_connection.index_remove("test")

    def test_index_remove_extra_args(self):
        # pass 'ns', 'idx_name', 'policy', and an extra argument
        with pytest.raises(TypeError):
            self.as_connection.index_remove("test", "demo", {}, "index_name")

    @pytest.mark.parametrize(
        "ns, idx_name, policy",
        (("test", "idx", "policy"), ("test", 5, {}), (5, "idx", {}), ("test", None, {}), (None, "idx", {})),
    )
    def test_index_remove_wrong_arg_types(self, ns, idx_name, policy):
        with pytest.raises(e.ParamError):
            self.as_connection.index_remove(ns, idx_name, policy)



# -*- coding: utf-8 -*-


import pytest

import aerospike

from aerospike_helpers import expressions as exp
from aerospike import exception as e
from .test_base_class import TestBaseClass
from .as_status_codes import AerospikeStatus


def add_indexes_and_udfs(client):
    """
    Load the UDFs used in the tests and setup indexes
    """
    policy = {}
    try:
        client.index_integer_create("test", "demo", "age", "age_index", policy)
    except e.IndexFoundError:
        pass
    try:
        client.index_integer_create("test", "demo", "age1", "age_index1", policy)
    except e.IndexFoundError:
        pass

    udf_type = 0
    udf_files = ("sample.lua", "test_record_udf.lua", "udf_basic_ops.lua")

    for module in udf_files:
        client.udf_put(module, udf_type, policy)


def remove_indexes_and_udfs(client):
    """
    Remove all of the UDFS and indexes created for these tests
    """
    policy = {}

    try:
        client.index_remove("test", "age_index", policy)
    except e.IndexNotFound:
        pass

    try:
        client.index_remove("test", "age_index1", policy)
    except e.IndexNotFound:
        pass

    udf_files = ("sample.lua", "test_record_udf.lua", "udf_basic_ops.lua")

    for module in udf_files:
        client.udf_remove(module, policy)


class TestBatchApply(TestBaseClass):
    def setup_class(cls):
        # Register setup and teardown functions
        cls.connection_setup_functions = [add_indexes_and_udfs]
        cls.connection_teardown_functions = [remove_indexes_and_udfs]

    @pytest.fixture(autouse=True)
    def setup(self, request, connection_with_config_funcs):
        as_connection = connection_with_config_funcs

        if self.server_version < [6, 0]:
            pytest.mark.xfail(reason="Servers older than 6.0 do not support batch apply.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"
        self.keys = []
        self.batch_size = 5

        for i in range(self.batch_size):
            key = ("test", "demo", i)
            rec = {
                "name": ["name%s" % (str(i))],
                "addr": "name%s" % (str(i)),
                "age": i,
                "no": i,
                "basic_map": {"k30": 6, "k20": 5, "k10": 1},
            }
            as_connection.put(key, rec)
            self.keys.append(key)

        def teardown():
            for i in range(self.batch_size):
                key = ("test", "demo", i)

                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    continue

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "name, keys, module, function, args, policy_batch, policy_batch_apply, exp_res, exp_rec, exp_change, \
            change_bin",
        [
            (
                "list-append",
                [("test", "demo", 0), ("test", "demo", 2), ("test", "demo", 4)],
                "sample",
                "list_append",
                ["name", 1],
                None,
                None,
                [AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK, AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": 0}, {"SUCCESS": 0}, {"SUCCESS": 0}],
                [
                    ["name0", 1],
                    ["name2", 1],
                    ["name4", 1],
                ],
                "name",
            ),
            (
                "return-bool",
                [("test", "demo", 0)],
                "test_record_udf",
                "bool_check",
                [],
                {},
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": True}],
                [],
                None,
            ),
            (
                "map-iterate",
                [("test", "demo", 0)],
                "test_record_udf",
                "map_iterate",
                ["basic_map", 555],
                {},
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": None}],
                [{"k30": 555, "k20": 555, "k10": 555}],
                "basic_map",
            ),
            (
                "return-bool-policy-batch",
                [("test", "demo", 0)],
                "test_record_udf",
                "bool_check",
                [],
                {
                    "max_retries": 2,
                    "allow_inline_ssd": True,
                    "respond_all_keys": False,
                    "expressions": exp.Eq(
                        exp.ListGetByRank(
                            None, aerospike.LIST_RETURN_VALUE, exp.ResultType.STRING, 0, exp.ListBin("name")
                        ),
                        "name0",
                    ).compile(),
                },
                {},
                [AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": True}],
                [],
                None,
            ),
            (
                "return-bool-policy-batch-apply",
                [("test", "demo", 0)],
                "test_record_udf",
                "bool_check",
                [],
                {},
                {
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                    "ttl": aerospike.TTL_DONT_UPDATE,
                    "durable_delete": False,
                    "expressions": exp.Eq(exp.IntBin("no"), 0).compile(),
                },
                [AerospikeStatus.AEROSPIKE_OK],
                [{"SUCCESS": True}],
                [],
                None,
            ),
        ],
    )
    def test_batch_apply_pos(
        self,
        name,
        keys,
        module,
        function,
        args,
        policy_batch,
        policy_batch_apply,
        exp_res,
        exp_rec,
        exp_change,
        change_bin,
    ):
        """
        Test batch_apply positive.
        """

        res = self.as_connection.batch_apply(keys, module, function, args, policy_batch, policy_batch_apply)

        for i, batch_rec in enumerate(res.batch_records):
            assert batch_rec.result == exp_res[i]
            assert batch_rec.key[:3] == keys[i]  # checking key
            assert batch_rec.record[2] == exp_rec[i]  # checking bins
            assert batch_rec.record[0][:3] == keys[i]  # checking key in record

            if exp_change:
                changed_rec = self.as_connection.get(keys[i])
                assert changed_rec[2][change_bin] == exp_change[i]

    def test_batch_apply_many_pos(self):
        """
        Test batch operate with many keys.
        """

        keys = [("test", "demo", i) for i in range(100, 1000)]
        module = "sample"
        func = "list_append"
        args = ["name", 10]
        policy_batch = {}
        policy_batch_apply = {}

        try:
            for i, key in enumerate(keys):
                self.as_connection.put(key, {"name": ["name" + str(i)]})

            res = self.as_connection.batch_apply(keys, module, func, args, policy_batch, policy_batch_apply)

            for i, batch_rec in enumerate(res.batch_records):
                assert batch_rec.result == AerospikeStatus.AEROSPIKE_OK
                assert batch_rec.key[:3] == keys[i]  # checking key
                assert batch_rec.record[2] == {"SUCCESS": 0}  # checking bins
                assert batch_rec.record[0][:3] == keys[i]  # checking key in record

            res = self.as_connection.get_many(keys)
            for i, rec in enumerate(res):
                assert rec[2]["name"] == ["name" + str(i), 10]

        except Exception as ex:
            raise (ex)
        finally:
            for key in keys:
                try:
                    self.as_connection.remove(key)
                except e.RecordNotFound:
                    continue

    @pytest.mark.parametrize(
        "name, keys, module, function, args, policy_batch, policy_batch_apply, exp_res",
        [
            (
                "bad-key",
                [("bad-key", i) for i in range(1000)],
                "sample",
                "list_append",
                ["name", 1],
                {},
                {},
                e.ParamError,
            ),
            ("bad-module", [("test", "demo", 1)], {"bad": "mod"}, "list_append", ["name", 1], {}, {}, e.ParamError),
            # ( NOTE These aren't raised but should be reflected in BatchRecords.result
            #     "non-existent-module",
            #     [
            #         ("test", "demo", 1)
            #     ],
            #     "fake_mod",
            #     "list_append",
            #     ["name", 1],
            #     {},
            #     {},
            #     e.AerospikeError
            # ),
            ("bad-func", [("test", "demo", 1)], "sample", {"bad": "func"}, ["name", 1], {}, {}, e.ParamError),
            # (
            #     "non-existent-func",
            #     [
            #         ("test", "demo", 1)
            #     ],
            #     "sample",
            #     "fake-func",
            #     ["name", 1],
            #     {},
            #     {},
            #     e.AerospikeError
            # ),
            ("bad-args", [("test", "demo", 1)], "sample", "list_append", {"bad": "args"}, {}, {}, e.ParamError),
            (
                "bad-batch-policy",
                [("test", "demo", 1)],
                "sample",
                "list_append",
                ["name", 1],
                ["bad-batch-policy"],
                {},
                e.ParamError,
            ),
            (
                "bad-batch-apply-policy",
                [("test", "demo", 1)],
                "sample",
                "list_append",
                ["name", 1],
                {},
                ["bad-batch-apply-policy"],
                e.ParamError,
            ),
        ],
    )
    def test_batch_apply_neg(self, name, keys, module, function, args, policy_batch, policy_batch_apply, exp_res):
        """
        Test batch_apply negative.
        """

        with pytest.raises(exp_res):
            self.as_connection.batch_apply(keys, module, function, args, policy_batch, policy_batch_apply)

    def test_batch_apply_neg_connection(self):
        """
        Test batch_apply negative with bad connection.
        """

        module = "lua_mod"
        function = "lua_func"
        args = []
        keys = []

        exp_res = e.ClientError

        with pytest.raises(exp_res):
            bad_client = aerospike.client({"hosts": [("bad_addr", 3000)]})
            bad_client.batch_apply(keys, module, function, args)



# -*- coding: utf-8 -*-
import pytest

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e


@pytest.mark.usefixtures("as_connection")
class TestRemovebin(object):
    @pytest.mark.parametrize(
        "key, record, bin_for_removal",
        [
            (("test", "demo", 1), {"Name": "Steve", "age": 60}, ["age"]),
            (("test", "demo", 2), {"name": "jeff", "age": 45}, ["name"]),
        ],
    )
    def test_pos_remove_bin_with_correct_parameters(self, key, record, bin_for_removal, put_data):
        """
        Invoke remove_bin() with correct parameters
        """
        put_data(self.as_connection, key, record)
        self.as_connection.remove_bin(key, bin_for_removal)

        (key, _, bins) = self.as_connection.get(key)

        del record["".join(bin_for_removal)]
        assert bins == record

    def test_pos_remove_bin_with_correct_policy(self, put_data):
        """
        Invoke remove_bin() with correct policy
        """
        key = ("test", "demo", 1)
        record = {"Name": "Herry", "age": 60}
        policy = {"timeout": 180000}
        put_data(self.as_connection, key, record)
        self.as_connection.remove_bin(key, ["age"], {}, policy)

        (key, _, bins) = self.as_connection.get(key)
        del record["age"]
        assert bins == record

    def test_pos_remove_bin_with_policy_send_gen_ignore(self, put_data):
        """
        Invoke remove_bin() with policy send
        """
        key = ("test", "demo", "policy_send_gen_ingnore")
        record = {"Name": "Herry", "age": 60, "address": "202, washingtoon"}
        put_data(self.as_connection, key, record)

        policy = {
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }
        meta = {"gen": 2, "ttl": 1000}
        self.as_connection.remove_bin(key, ["age"], meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        del record["age"]
        assert bins == record
        assert key == ("test", "demo", None, bytearray(b"\xbd\x87-\x84\xae99|\x06z\x12\xf3\xef\x12\xb9\x1a\xa2\x1a;'"))

    def test_pos_remove_bin_with_policy_send_gen_eq_positive(self, put_data):
        """
        Invoke remove_bin() with policy gen eq less
        """
        key = ("test", "demo", 1)
        record = {"Company": "Apple", "years": 30, "address": "202, sillicon Vally"}
        put_data(self.as_connection, key, record)
        policy = {
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_EQ,
        }

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1000}

        self.as_connection.remove_bin(key, ["years"], meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        del record["years"]
        assert bins == record
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_pos_remove_bin_with_policy_key_digest(self, put_data):
        """
        Invoke remove_bin() with policy key digest
        """
        key = ("test", "demo", None, bytearray("asd;as[d'as;djk;uyfl", "utf-8"))
        record = {"age": 1, "name": "name1"}

        put_data(self.as_connection, key, record)
        policy = {"key": aerospike.POLICY_KEY_DIGEST}
        self.as_connection.remove_bin(key, ["age"], {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        del record["age"]
        assert bins == record
        assert key == ("test", "demo", None, bytearray(b"asd;as[d'as;djk;uyfl"))

    def test_pos_remove_bin_with_single_bin_in_a_record(self):
        """
        Invoke remove_bin() with policy key digest
        """
        key = ("test", "demo", "single-bin")
        try:
            self.as_connection.remove(key)
        except Exception:
            pass

        rec = {"name": "single"}
        self.as_connection.put(key, rec)

        policy = {}
        self.as_connection.remove_bin(key, ["name"], {}, policy)

        try:
            _, _, bins = self.as_connection.get(key)
            assert bins is None
        except e.RecordNotFound as exception:
            assert exception.code == 2

    def test_pos_remove_bin_no_bin(self, put_data):
        """
        Invoke remove_bin() no bin
        """
        key = ("test", "demo", 1)
        record = {"name": "jeff", "age": 45}
        put_data(self.as_connection, key, record)
        try:
            self.as_connection.remove_bin(key, [])
            (key, _, bins) = self.as_connection.get(key)
            assert bins == record
        except e.InvalidRequest:
            pass

    @pytest.mark.parametrize(
        "key, record, bins_for_removal",
        [
            (("test", "demo", 1), {"name": "Devid", "age": 30}, ["name", "age"]),
            (("test", "demo", 3), {"name": "jeff", "age": 45}, ["name", "age"]),
            (("test", "demo", 4), {"name": "jeff", "age": 45}, ["name", "age"]),
        ],
    )
    def test_pos_remove_bin_with_unicode_all(self, key, record, bins_for_removal, put_data):
        """
        Invoke remove_bin() with unicode bin name
        """
        put_data(self.as_connection, key, record)
        self.as_connection.remove_bin(key, bins_for_removal)

        try:
            (key, _, _) = self.as_connection.get(key)

        except e.RecordNotFound as exception:
            assert exception.code == 2

    @pytest.mark.parametrize(
        "key, record, policy, bin_for_removal",
        [
            (
                ("test", "demo", "p_commit_level_all"),
                {"Name": "John", "age": 30, "address": "202, washingtoon"},
                {
                    "retry": aerospike.POLICY_RETRY_ONCE,
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit": aerospike.POLICY_COMMIT_LEVEL_ALL,
                },
                "age",
            ),
            (
                ("test", "demo", "p_commit_level_master"),
                {"Name": "John", "age": 30, "address": "202, washingtoon"},
                {
                    "retry": aerospike.POLICY_RETRY_ONCE,
                    "key": aerospike.POLICY_KEY_SEND,
                    "commit": aerospike.POLICY_COMMIT_LEVEL_MASTER,
                },
                "age",
            ),
            (
                ("test", "demo", "p_gen_GT"),
                {"Name": "John", "age": 30, "address": "202, washingtoon"},
                {
                    "retry": aerospike.POLICY_RETRY_ONCE,
                    "key": aerospike.POLICY_KEY_SEND,
                    "gen": aerospike.POLICY_GEN_GT,
                },
                "age",
            ),
        ],
    )
    def test_pos_remove_bin_with_policy(self, key, record, policy, bin_for_removal, put_data):
        """
        Invoke remove_bin() with policy
        """
        key_digest = self.as_connection.get_key_digest(key[0], key[1], key[2])

        put_data(self.as_connection, key, record)
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1000}

        self.as_connection.remove_bin(key, [bin_for_removal], meta, policy)

        (key, meta, bins) = self.as_connection.get(key)

        del record[bin_for_removal]

        assert bins == record
        assert key == ("test", "demo", None, key_digest)

    # Negative Tests

    @pytest.mark.parametrize(
        "key, bin_for_removal, ex_code, ex_msg",
        [
            # key_is_none
            (None, ["age"], -2, "key is invalid"),
            (("test", "demo", 1), None, -2, "Bins should be a list"),  # bin_is_none
        ],
    )
    def test_neg_remove_bin_with_none(self, key, bin_for_removal, ex_code, ex_msg):
        """
        Invoke remove_bin() with none
        """
        try:
            self.as_connection.remove_bin(None, bin_for_removal)

        except e.ParamError as exception:
            assert exception.code == ex_code
            assert exception.msg == ex_msg

    def test_neg_remove_bin_with_correct_parameters_without_connection(self):
        """
        Invoke remove_bin() with correct parameters without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        key = ("test", "demo", 1)

        try:
            client1.remove_bin(key, ["age"])

        except e.ClusterError as exception:
            assert exception.code == 11

    def test_neg_remove_bin_with_incorrect_meta(self):
        """
        Invoke remove_bin() with incorrect meta
        """
        key = ("test", "demo", 1)
        policy = {
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_IGNORE,
        }
        try:
            self.as_connection.remove_bin(key, ["age"], policy)

        except (e.ClusterError, e.RecordNotFound):
            pass

    def test_neg_remove_bin_with_incorrect_policy(self):
        """
        Invoke remove_bin() with incorrect policy
        """
        key = ("test", "demo", 1)

        policy = {"time": 1001}
        try:
            self.as_connection.remove_bin(key, ["age"], {}, policy)

        except (e.ClientError, e.RecordNotFound):
            pass

    def test_neg_remove_bin_with_no_parameters(self):
        """
        Invoke remove_bin() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.remove_bin()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_remove_bin_with_policy_send_gen_eq_not_equal(self, put_data):
        """
        Invoke remove_bin() with policy gen eq not equal
        """
        key = ("test", "demo", 1)
        record = {"Name": "John", "age": 30, "address": "202, washingtoon"}
        put_data(self.as_connection, key, record)

        policy = {
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_EQ,
        }
        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen + 5, "ttl": 1000}

        try:
            self.as_connection.remove_bin(key, ["age"], meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == record
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\xb7\xf4\xb88\x89\xe2\xdag\xdeh>\x1d\xf6\x91\x9a\x1e\xac\xc4F\xc8"),
        )

    def test_neg_remove_bin_with_policy_send_gen_GT_lesser(self, put_data):
        """
        Invoke remove_bin() with policy gen GT lesser
        """
        key = ("test", "demo", "send_gen_GT_lesser")
        record = {"Name": "John", "age": 30, "address": "202, washingtoon"}
        put_data(self.as_connection, key, record)
        policy = {
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
            "gen": aerospike.POLICY_GEN_GT,
        }

        (key, meta) = self.as_connection.exists(key)
        gen = meta["gen"]
        meta = {"gen": gen, "ttl": 1000}

        try:
            self.as_connection.remove_bin(key, ["age"], meta, policy)

        except e.RecordGenerationError as exception:
            assert exception.code == 3

        (key, meta, bins) = self.as_connection.get(key)

        assert bins == record
        assert key == (
            "test",
            "demo",
            None,
            bytearray(b"\x1b\xb9\xda`\xa7\xbd\xe0\xc1\xdet1\xe4\x82\x94\xc7\xb3\xd8\xd5\x7f."),
        )

    def test_neg_remove_bin_with_incorrect_policy_value(self):
        """
        Invoke remove_bin() with incorrect policy value
        """
        key = ("test", "demo", 1)

        policy = {"total_timeout": 0.5}
        try:
            self.as_connection.remove_bin(key, ["age"], {}, policy)

        except e.ClientError as exception:
            assert exception.code == -1

    @pytest.mark.parametrize(
        "key, bin_for_removal, ex_code",
        [
            (("test", "demo", 1), ["non-existent"], 0),  # non-existent bin
            (("test", "demo", "non-existent"), ["age"], 0),  # non-existent key
        ],
    )
    def test_neg_remove_bin_with_nonexistent_data(self, key, bin_for_removal, ex_code):
        """
        Invoke remove_bin() with non-existent data
        """
        try:
            self.as_connection.remove_bin(key, bin_for_removal)

        except e.RecordNotFound:
            pass

    def test_neg_remove_bin_with_extra_parameter(self):
        """
        Invoke remove_bin() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.remove_bin(key, ["age"], {}, policy, "")

        assert "remove_bin() takes at most 4 arguments (5 given)" in str(typeError.value)

    def test_remove_bin_with_non_string_bin_nane(self):
        """
        Invoke remove_bin() with extra parameter.
        """
        key = ("test", "demo", 1)
        with pytest.raises(e.ClientError):
            self.as_connection.remove_bin(key, [1.5])

    @pytest.mark.skip(reason="system error")
    def test_remove_bin_with_ttl_too_large(self, put_data):
        key = ("test", "demo", 1)
        record = {"Name": "Herry", "age": 60}
        put_data(self.as_connection, key, record)
        meta = {"gen": 2, "ttl": 2**65}
        with pytest.raises(e.ClientError):
            self.as_connection.remove_bin(key, ["age"], meta=meta)

    @pytest.mark.skip(reason="system error")
    def test_remove_bin_with_gen_too_large(self, put_data):
        key = ("test", "demo", 1)
        record = {"Name": "Herry", "age": 60}
        put_data(self.as_connection, key, record)
        meta = {"gen": 2**65, "ttl": 2}
        with pytest.raises(e.ClientError):
            self.as_connection.remove_bin(key, ["age"], meta=meta)



# Callback function
# Calculates new elo for a player
def updateElo(record):
    keyTuple, _, bins = record
    # Add score to elo
    bins["elo"] = bins["elo"] + bins["score"]
    client.put(keyTuple, bins)

query.foreach(updateElo)

# Player elos should be updated
records = client.get_many(keyTuples)
for _, _, bins in records:
    print(bins)
# {'score': 100, 'elo': 1500}
# {'score': 20, 'elo': 1520}
# {'score': 10, 'elo': 1110}
# {'score': 200, 'elo': 1100}



import os

try:
    import ConfigParser as configparser
except Exception:
    import configparser
import aerospike


class TestBaseClass(object):
    hostlist = []
    user = None
    password = None
    using_tls = False
    using_auth = False
    should_xfail = False
    using_enterprise = False
    major_ver = 0
    minor_ver = 0

    @staticmethod
    def get_hosts():
        config = configparser.ConfigParser()
        config.read("config.conf")
        if config.has_option("enterprise-edition", "hosts"):
            hosts_str = config.get("enterprise-edition", "hosts")
            if hosts_str != "None":
                TestBaseClass.hostlist = TestBaseClass.parse_hosts(hosts_str)
                if TestBaseClass.hostlist:
                    TestBaseClass.using_enterprise = True
        if len(TestBaseClass.hostlist) > 0:
            if config.has_option("enterprise-edition", "user"):
                TestBaseClass.user = config.get("enterprise-edition", "user")
            if config.has_option("enterprise-edition", "password"):
                TestBaseClass.password = config.get("enterprise-edition", "password")

                # If the password is empty, assume we aren't using authentication
                if not TestBaseClass.password.strip():
                    TestBaseClass.password = None
                    TestBaseClass.user = None
                else:
                    TestBaseClass.using_auth = True
        else:
            TestBaseClass.hostlist = TestBaseClass.parse_hosts(config.get("community-edition", "hosts"))

        return TestBaseClass.hostlist

    @staticmethod
    def temporary_xfail():
        if TestBaseClass.should_xfail:
            return True
        else:
            env_val = os.environ.get("TEMPORARY_XFAIL")
            TestBaseClass.should_xfail = bool(env_val)
        return TestBaseClass.should_xfail

    @staticmethod
    def get_tls_info():
        tls_dict = {}
        config = configparser.ConfigParser()
        config.read("config.conf")

        if config.has_option("tls", "enable") and (
            config.get("tls", "enable") != "" and config.get("tls", "enable") != "0"
        ):
            TestBaseClass.using_tls = True
            tls_dict["enable"] = True

            if config.has_option("tls", "cafile"):
                if config.get("tls", "cafile") != "":
                    tls_dict["cafile"] = config.get("tls", "cafile")

            if config.has_option("tls", "capath"):
                if config.get("tls", "capath") != "":
                    tls_dict["capath"] = config.get("tls", "capath")

            if config.has_option("tls", "protocols"):
                if config.get("tls", "protocols") != "":
                    tls_dict["protocols"] = config.get("tls", "protocols")

            if config.has_option("tls", "cipher_suite"):
                if config.get("tls", "cipher_suite") != "":
                    tls_dict["cipher_suite"] = config.get("tls", "cipher_suite")

            if config.has_option("tls", "keyfile"):
                if config.get("tls", "keyfile") != "":
                    tls_dict["keyfile"] = config.get("tls", "keyfile")

            if config.has_option("tls", "cert_blacklist"):
                if config.get("tls", "cert_blacklist") != "":
                    tls_dict["cert_blacklist"] = config.get("tls", "cert_blacklist")

            if config.has_option("tls", "certfile"):
                if config.get("tls", "certfile") != "":
                    tls_dict["certfile"] = config.get("tls", "certfile")

            if config.has_option("tls", "crl_check"):
                if config.get("tls", "crl_check") == "True":
                    tls_dict["crl_check"] = True

            if config.has_option("tls", "crl_check_all"):
                if config.get("tls", "crl_check_all") == "True":
                    tls_dict["crl_check_all"] = True

            if config.has_option("tls", "log_session_info"):
                if config.get("tls", "log_session_info") == "True":
                    tls_dict["log_session_info"] = True

            if config.has_option("tls", "max_socket_idle"):
                tls_dict["max_socket_idle"] = int(config.get("tls", "max_socket_idle"))
        else:
            return tls_dict
        return tls_dict

    @staticmethod
    def get_policies_info():
        policies_dict = {}
        config = configparser.ConfigParser()
        config.read("config.conf")

        policies_dict["auth_mode"] = int(0)
        if config.has_option("policies", "auth_mode"):
            auth_mode = config.get("policies", "auth_mode")
            if auth_mode != "None" and auth_mode.isdigit():
                policies_dict["auth_mode"] = int(auth_mode)

        return policies_dict

    @staticmethod
    def parse_hosts(host_str):
        hosts = []
        host_str = host_str.strip()

        hlist = host_str.split(";")
        for host in hlist:
            tls_name = None
            if host.rfind("|") != -1:
                index = host.rfind("|")
                tls_name = host[index + 1 :]
                host = host[:index]

            host = host.strip()
            last_colon = host.rfind(":")
            if last_colon != -1:
                host = [host[:last_colon], host[last_colon + 1 :]]
            else:
                host = [host]
            if len(host) == 2:
                host[1] = int(host[1])
                if tls_name:
                    host.append(tls_name)
                hosts.append(tuple(host))
        return hosts

    @staticmethod
    def get_new_connection(add_config=None):
        """
        Merge the add_config dict with that in the default configuration
        and return a newly connected client
        """
        add_config = add_config if add_config else {}
        config = TestBaseClass.get_connection_config()
        for key in add_config:
            config[key] = add_config[key]

        client = aerospike.client(config)
        if config["user"] is None and config["password"] is None:
            client.connect()
        else:
            client.connect(config["user"], config["password"])

        if client is not None:
            build_info = client.info_all("build")
            res = []
            for _, (error, result) in list(build_info.items()):
                res = None if error is not None else result.strip().strip(";").strip(":")
                res = None if res is None or len(res) == 0 else res
                if res is not None:
                    break
            res = res.split(".")
            # major_ver = res[0]
            # minor_ver = res[1]
            # print("major_ver:", major_ver, "minor_ver:", minor_ver)

        return client

    @staticmethod
    def tls_in_use():
        if TestBaseClass.using_tls:
            return True
        else:
            return True

    @staticmethod
    def auth_in_use():
        if TestBaseClass.using_auth:
            return True
        else:
            return False

    @staticmethod
    def enterprise_in_use():
        if TestBaseClass.using_enterprise:
            return True
        else:
            return False

    @staticmethod
    def get_connection_config():
        config = {}
        hosts_conf = TestBaseClass.get_hosts()
        tls_conf = TestBaseClass.get_tls_info()
        policies_conf = TestBaseClass.get_policies_info()
        config["hosts"] = hosts_conf
        config["tls"] = tls_conf
        config["policies"] = policies_conf
        config["user"] = TestBaseClass.user
        config["password"] = TestBaseClass.password

        # Disable total_timeout and timeout
        # config["timeout"] = 0
        # config["total_timeout"] = 0
        # Also have to set for every operation policy since
        # total_timeout at the root level is deprecated
        policy_names = ("read", "write", "apply", "operate", "remove", "query", "scan", "batch")
        for policy_name in policy_names:
            config["policies"][policy_name] = {}
            # 3 minutes
            # Override server's total (transaction) timeout
            config["policies"][policy_name]["total_timeout"] = 180000
            # Must hear back from server after a certain number of seconds
            config["policies"][policy_name]["socket_timeout"] = 180000
        # config["max_socket_idle"] = 60

        config["policies"]["info"] = {}
        config["policies"]["info"]["timeout"] = 180000

        return config



# -*- coding: utf-8 -*-

import pytest
import asyncio

try:
    import cPickle as pickle
except Exception:
    import pickle
from . import test_data

# from collections import OrderedDict

from .test_base_class import TestBaseClass
import aerospike
from aerospike import exception as e
from aerospike_helpers.awaitable import io


class SomeClass(object):
    pass


aerospike.init_async()


@pytest.mark.usefixtures("as_connection")
class TestPut:
    @pytest.mark.asyncio
    async def test_pos_put_with_policy_exists_create_or_replace(self):
        """
        Invoke put() for a record with create or replace policy positive
        If record exists and will be replaced
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_CREATE_OR_REPLACE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_CREATE_OR_REPLACE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }
        assert 0 == self.as_connection.put(key, rec, meta, policy)

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_put_with_policy_exists_ignore_create(self):
        """
        Invoke put() for a record with ignore.
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_IGNORE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_put_with_policy_exists_ignore_update(self):
        """
        Invoke put() for a record with ignore.
        Ignore for an existing records should update without any error
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_IGNORE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins

        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_IGNORE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_put_with_policy_replace(self):
        """
        Invoke put() for a record with replace positive.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {"exists": aerospike.POLICY_EXISTS_REPLACE}

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_put_with_policy_exists_update_positive(self):
        """
        Invoke put() for a record with all policies.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {"exists": aerospike.POLICY_EXISTS_UPDATE}

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_put_with_policy_gen_GT(self):
        """
        Invoke put() for a record with generation as GT positive
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins
        gen = meta["gen"]
        assert gen == 1
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_GT}
        meta = {"gen": gen + 5}

        async def async_io(key=None, rec=None, meta=None, policy=None):
            await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_pos_put_with_policy_gen_ignore(self):
        """
        Invoke put() for a record with generation as gen_ignore
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert rec == bins

        gen = meta["gen"]
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_IGNORE}
        meta = {"gen": gen}

        async def async_io(key=None, rec=None, meta=None, policy=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert rec == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, record, meta, policy",
        [
            (("test", "demo", 1), {"name": "john"}, {"gen": True, "ttl": 25000}, {}),
            (("test", "demo", 1), {"name": "john"}, {"gen": 3, "ttl": True}, {}),
            (("test", "demo", 1), {"name": "john"}, {"gen": True, "ttl": True}, {}),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": True, "ttl": aerospike.TTL_NAMESPACE_DEFAULT},
                {},
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": True, "ttl": aerospike.TTL_NEVER_EXPIRE},
                {},
            ),
        ],
    )
    async def test_pos_put_with_metadata_bool(self, key, record, meta, policy, put_data):
        """
        Invoke put() for a record with generation as boolean.
        """
        put_data(self.as_connection, key, record, meta, policy)

        async def async_io(key=None, policy=None):
            (key, meta, bins) = await io.get(self.as_connection, key)
            assert bins == record
            assert meta["gen"] is not None
            assert meta["ttl"] is not None

        await asyncio.gather(async_io(key))

    @pytest.mark.asyncio
    async def test_pos_put_user_serializer_no_deserializer(self):
        """
        Invoke put() for float data record with user serializer is
        registered, but deserializer is not registered.
        """
        key = ("test", "demo", "put_user_serializer")

        rec = {"pi": 3.14}

        def serialize_function(val):
            return pickle.dumps(val)

        aerospike.set_serializer(serialize_function)

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, {}, {}, aerospike.SERIALIZER_USER))

        _, _, bins = self.as_connection.get(key)

        if self.skip_old_server is False:
            assert bins == {"pi": 3.14}
        else:
            assert bins == {"pi": bytearray(b"F3.1400000000000001\n.")}

        self.as_connection.remove(key)

    # put negative
    @pytest.mark.asyncio
    async def test_neg_put_with_no_parameters(self):
        """
        Invoke put() without any parameters.
        """

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.ParamError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io())

    @pytest.mark.asyncio
    async def test_neg_put_without_record(self):
        """
        Invoke put() without any record data.
        """
        key = ("test", "demo", 1)

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.ParamError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key))

    @pytest.mark.asyncio
    @pytest.mark.parametrize("key, ex_code, ex_msg", test_data.key_neg)
    async def test_neg_put_with_none(self, key, ex_code, ex_msg, record={}):
        """
        Invoke put() with invalid data
        """

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.ParamError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, record))

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, ex_code, ex_msg, record",
        [
            (("test", "demo", None), -2, "either key or digest is required", {"name": "John"}),
        ],
    )
    async def test_neg_put_with_invalid_record(self, key, ex_code, ex_msg, record):
        """
        Invoke put() with invalid data
        """

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.ParamError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, record))

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "key, record, exception_code",
        [
            # Non-existing NS & Set
            (("demo", "test", 1), {"a": ["!@#!#$%#", bytearray("ASD@#$AR#$@#ERQ#", "utf-8")]}, 20),
            # Non-existing Namespace
            (("test1", "demo", 1), {"i": "asdadasd"}, 20),
        ],
    )
    async def test_neg_put_with_wrong_ns_and_set(self, key, record, exception_code):
        """
        Invoke put() with non-existent data
        """

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.ClientError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, record))

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_gen_EQ_less(self):
        """
        Invoke put() for a record with generation as EQ less
        """
        key = ("test", "demo", "policy_gen_EQ_key")

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_EQ}
        meta = {"gen": 10}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.RecordGenerationError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins

        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_gen_EQ_more(self):
        """
        Invoke put() for a record with generation as EQ less
        """
        key = ("test", "demo", "policy_gen_EQ_more_key")

        rec = {"name": "John"}
        meta = {"gen": 10, "ttl": 25000}
        policy = {}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_EQ}
        meta = {"gen": 4}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.RecordGenerationError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins

        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_exists_create(self):
        """
        Invoke put() for a record with all policies.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins

        rec = {"name": "Smith"}
        policy = {"exists": aerospike.POLICY_EXISTS_CREATE}
        meta = {"gen": 2}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.RecordExistsError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins

        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_exists_replace_negative(self):
        """
        Invoke put() for a record with replace policy negative.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_REPLACE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.RecordNotFound):
                assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_replace(self):
        """
        Invoke put() for a record with replace positive.
        """
        key = ("test", "demo", 1)

        rec = {"name": "Smith"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {"exists": aerospike.POLICY_EXISTS_REPLACE}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.RecordNotFound):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_exists_update_negative(self):
        """
        Invoke put() for a record with update policy negative.
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {
            "exists": aerospike.POLICY_EXISTS_UPDATE,
            "gen": aerospike.POLICY_GEN_IGNORE,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "key": aerospike.POLICY_KEY_SEND,
        }

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            try:
                assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)
            except e.RecordNotFound as exception:
                assert exception.code == 2
                assert exception.msg == "AEROSPIKE_ERR_RECORD_NOT_FOUND"

        await asyncio.gather(async_io(key, rec, meta, policy))

    @pytest.mark.asyncio
    async def test_neg_put_with_policy_gen_GT_lesser(self):
        """
        Invoke put() for a record with generation as GT lesser
        """
        key = ("test", "demo", 1)

        rec = {"name": "John"}
        meta = {"gen": 2, "ttl": 25000}
        policy = {}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)

        assert {"name": "John"} == bins
        gen = meta["gen"]
        rec = {"name": "Smith"}
        policy = {"gen": aerospike.POLICY_GEN_GT}
        meta = {"gen": gen}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            try:
                assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)
            except e.RecordGenerationError as exception:
                assert exception.code == 3
                assert exception.msg == "AEROSPIKE_ERR_RECORD_GENERATION"

        await asyncio.gather(async_io(key, rec, meta, policy))

        (key, meta, bins) = self.as_connection.get(key)
        assert {"name": "John"} == bins
        self.as_connection.remove(key)

    @pytest.mark.asyncio
    async def test_neg_put_with_string_record_without_connection(self):
        """
        Invoke put() for a record with string data without connection
        """
        config = TestBaseClass.get_connection_config()
        client1 = aerospike.client(config)
        client1.close()

        key = ("test", "demo", 1)

        bins = {"name": "John"}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            try:
                assert 0 == await io.put(client1, key, rec, meta, policy, serialize)
            except e.ClusterError as exception:
                assert exception.code == 11

        await asyncio.gather(async_io(key, bins))

    @pytest.mark.parametrize(
        "key, record, meta, policy, ex_code, ex_msg",
        [
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": "wrong", "ttl": 25000},
                {},  # Gen as string
                -2,
                "Generation should be an int or long",
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": 3, "ttl": "25000"},
                {},  # ttl as string
                -2,
                "TTL should be an int or long",
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},
                {"gen": 3, "ttl": 25000},
                {"total_timeout": "1000"},  # Timeout as string
                -2,
                "timeout is invalid",
            ),
            (
                ("test", "demo", 1),
                {"name": "john"},  # Policy as string
                {"gen": 3, "ttl": 25000},
                "Policy",
                -2,
                "policy must be a dict",
            ),
            (
                ("test", "demo", 1),
                {"i": 13},  # Meta as string
                "OK",
                {},
                -2,
                "meta must be a dict",
            ),
            (
                ("test", "demo", 1),
                {"i": 13},  # Meta as string
                1234,
                {},
                -2,
                "meta must be a dict",
            ),
        ],
    )
    def test_neg_put_with_invalid_metadata(self, key, record, meta, policy, ex_code, ex_msg, put_data):
        """
        Invoke put() for a record with generation as string
        """
        with pytest.raises(e.ParamError):
            put_data(self.as_connection, key, record, meta, policy)
        #     # self.as_connection.remove(key)
        # except e.ParamError as exception:
        #     assert exception.code == ex_code
        #     assert exception.msg == ex_msg

    # put edge cases
    @pytest.mark.asyncio
    async def test_edge_put_record_with_bin_name_exceeding_max_limit(self):
        """
        Invoke put() with bin name exceeding the max limit of bin name.
        """
        key = ("test", "demo", "put_rec")
        put_record = {"a" * 50: "unimportant"}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.BinNameError):
                await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, put_record))

    @pytest.mark.asyncio
    async def test_edge_put_with_integer_greater_than_maxisze(self):
        """
        Invoke put() for a record with integer greater than max size
        """
        key = ("test", "demo", 1)

        bins = {"no": 111111111111111111111111111111111111111111111}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            try:
                assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)
            except e.ParamError as exception:
                assert exception.code == -2
                assert exception.msg == "integer value exceeds sys.maxsize"
            except SystemError:
                pass

        await asyncio.gather(async_io(key, bins))

    @pytest.mark.asyncio
    async def test_edge_put_with_key_as_an_integer_greater_than_maxsize(self):
        """
        Invoke put() for a record with integer greater than max size
        """
        key = ("test", "demo", 1111111111111111111111111111111111)

        bins = {"no": 11}

        async def async_io(key=None, rec=None, meta=None, policy=None, serialize=None):
            with pytest.raises(e.ParamError):
                assert 0 == await io.put(self.as_connection, key, rec, meta, policy, serialize)

        await asyncio.gather(async_io(key, bins))



# -*- coding: utf-8 -*-
##########################################################################
# Copyright 2013-2021 Aerospike, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##########################################################################


from __future__ import print_function
import aerospike
import json
import re
import sys
import os.path

from optparse import OptionParser
from aerospike import predicates as p

##########################################################################
# Option Parsing
##########################################################################

usage = "usage: %prog [options] where module function [args...]"

optparser = OptionParser(usage=usage, add_help_option=False)

optparser.add_option(
    "-U", "--username", dest="username", type="string", metavar="<USERNAME>",
    help="Username to connect to database.")

optparser.add_option(
    "-P", "--password", dest="password", type="string", metavar="<PASSWORD>",
    help="Password to connect to database.")

optparser.add_option(
    "-h", "--host", dest="host", type="string", default="127.0.0.1", metavar="<ADDRESS>",
    help="Address of Aerospike server.")

optparser.add_option(
    "-p", "--port", dest="port", type="int", default=3000, metavar="<PORT>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "--help", dest="help", action="store_true",
    help="Displays this message.")

optparser.add_option(
    "-n", "--namespace", dest="namespace", type="string", default="test", metavar="<NS>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-s", "--set", dest="set", type="string", default="demo", metavar="<SET>",
    help="Port of the Aerospike server.")

optparser.add_option(
    "-b", "--bins", dest="bins", type="string", action="append",
    help="Bins to select from each record.")

(options, args) = optparser.parse_args()

if options.help:
    optparser.print_help()
    print()
    sys.exit(1)

if len(args) < 3:
    optparser.print_help()
    print()
    sys.exit(1)

##########################################################################
# Client Configuration
##########################################################################

config = {
    'hosts': [(options.host, options.port)],
    'lua': {
        'user_path': os.path.dirname(__file__)
    }
}

##########################################################################
# Application
##########################################################################

exitCode = 0


def parse_arg(s):
    try:
        return json.loads(s)
    except ValueError:
        return s

try:

    # ----------------------------------------------------------------------------
    # Connect to Cluster
    # ----------------------------------------------------------------------------

    client = aerospike.client(config).connect(
        options.username, options.password)

    # ----------------------------------------------------------------------------
    # Perform Operation
    # ----------------------------------------------------------------------------

    try:

        re_bin = "(.{1,14})"
        re_str_eq = "\s+=\s*(?:(?:\"(.*)\")|(?:\'(.*)\'))"
        re_int_eq = "\s+=\s*(\d+)"
        re_int_rg = "\s+between\s+\(\s*(\d+)\s*,\s*(\d+)\s*\)"
        re_w = re.compile("%s(?:%s|%s|%s)" %
                          (re_bin, re_str_eq, re_int_eq, re_int_rg))

        namespace = options.namespace if options.namespace and options.namespace != 'None' else None
        set = options.set if options.set and options.set != 'None' else None

        args.reverse()
        where = args.pop()
        module = args.pop()
        function = args.pop()

        # If predicate is provided, then perform a query
        q = client.query(namespace, set)
        w = re_w.match(where)
        if w is not None:
            if w.group(2):
                b = w.group(1)
                v = w.group(2)
                q.where(p.equals(b, v))
            elif w.group(3):
                b = w.group(1)
                v = w.group(3)
                q.where(p.equals(b, v))
            elif w.group(4):
                b = w.group(1)
                v = int(w.group(4))
                q.where(p.equals(b, v))
            elif w.group(5) and w.group(6):
                b = w.group(1)
                l = int(w.group(5))
                u = int(w.group(6))
                q.where(p.between(b, l, u))

        if options.bins and len(options.bins) > 0:
            # project specified bins
            q.select(*options.bins)

        args.reverse()
        argl = list(map(parse_arg, args))
        print("argl == ", argl)
        q.apply(module, function, *argl)

        results = []

        # callback to be called for each record read
        def callback(result):
            results.append(result)
            print(result)

        # invoke the operations, and for each record invoke the callback
        q.foreach(callback)

        print("---")
        if len(results) == 1:
            print("OK, 1 result found.")
        else:
            print("OK, %d results found." % len(results))

    except Exception as eargs:
        print("error: {0}".format(eargs), file=sys.stderr)
        exitCode = 2

    # ----------------------------------------------------------------------------
    # Close Connection to Cluster
    # ----------------------------------------------------------------------------

    client.close()

except Exception as eargs:
    print("error: {0}".format(eargs), file=sys.stderr)
    exitCode = 3


##########################################################################
# Exit
##########################################################################

sys.exit(exitCode)



# -*- coding: utf-8 -*-
import pytest
import random
from aerospike import exception as e

import aerospike


class TestListExtend(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        keys = []
        for i in range(5):
            key = ("test", "demo", i)
            rec = {"name": "name%s" % (str(i)), "contact_no": [i, i + 1], "city": ["Pune", "Dehli"]}
            as_connection.put(key, rec)
            keys.append(key)

        def teardown():
            """
            Teardown method.
            """
            for key in keys:
                try:
                    as_connection.remove(key)
                except e.RecordNotFound:
                    pass

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, field, extend_value, expected",
        [
            (
                ("test", "demo", 1),  # extend the list with integer values
                "contact_no",
                [12, 56, 89],
                {"contact_no": [1, 2, 12, 56, 89], "name": "name1", "city": ["Pune", "Dehli"]},
            ),
            (
                ("test", "demo", 2),  # with float values
                "contact_no",
                [85.12, 85.46],
                {"contact_no": [2, 3, 85.12, 85.46], "city": ["Pune", "Dehli"], "name": "name2"},
            ),
            (
                ("test", "demo", 1),  # all values
                "contact_no",
                [False, [789, 45], 88, 15.2, "aa"],
                {"contact_no": [1, 2, 0, [789, 45], 88, 15.2, "aa"], "city": ["Pune", "Dehli"], "name": "name1"},
            ),
        ],
    )
    def test_pos_list_extend_with_list_of_values(self, key, field, extend_value, expected):
        """
        Invoke list_extend() extend the list with values
        """
        self.as_connection.list_extend(key, field, extend_value)
        (key, _, bins) = self.as_connection.get(key)
        assert bins == expected

    def test_pos_list_extend_with_policy(self):
        """
        Invoke list_extend() extend the list with integer values and
        policy is passed
        """
        key = ("test", "demo", 1)

        policy = {
            "timeout": 180000,
            "retry": aerospike.POLICY_RETRY_ONCE,
            "commit_level": aerospike.POLICY_COMMIT_LEVEL_MASTER,
        }

        self.as_connection.list_extend(key, "contact_no", [12, 56, 89], {}, policy)

        (key, _, bins) = self.as_connection.get(key)

        assert bins == {"contact_no": [1, 2, 12, 56, 89], "name": "name1", "city": ["Pune", "Dehli"]}

    def test_pos_list_extend_with_nonexistent_key(self):
        """
        Invoke list_extend() with non-existent key
        """
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 30
        length = random.randint(minLength, maxLength)
        key = ("test", "demo", "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com")
        status = self.as_connection.list_extend(key, "abc", [122, 789])
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)

        assert status == 0
        assert bins == {"abc": [122, 789]}

        self.as_connection.remove(key)

    def test_pos_list_extend_with_nonexistent_bin(self):
        """
        Invoke list_extend() with non-existent bin
        """
        key = ("test", "demo", 1)
        charSet = "abcdefghijklmnopqrstuvwxyz1234567890"
        minLength = 5
        maxLength = 10
        length = random.randint(minLength, maxLength)
        bin = "".join(map(lambda unused: random.choice(charSet), range(length))) + ".com"
        status = self.as_connection.list_extend(key, bin, [585, 789, 45])
        assert status == 0

        (key, _, bins) = self.as_connection.get(key)

        assert status == 0
        assert bins == {"contact_no": [1, 2], "name": "name1", "city": ["Pune", "Dehli"], bin: [585, 789, 45]}

    # Negative Tests
    def test_neg_list_extend_with_no_parameters(self):
        """
        Invoke list_extend() without any mandatory parameters.
        """
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_extend()
        assert "argument 'key' (pos 1)" in str(typeError.value)

    def test_neg_list_extend_with_incorrect_policy(self):
        """
        Invoke list_extend() with incorrect policy
        """
        key = ("test", "demo", 1)
        policy = {"timeout": 0.5}
        try:
            self.as_connection.list_extend(key, "contact_no", ["str"], {}, policy)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "timeout is invalid"

    def test_neg_list_extend_with_extra_parameter(self):
        """
        Invoke list_extend() with extra parameter.
        """
        key = ("test", "demo", 1)
        policy = {}
        with pytest.raises(TypeError) as typeError:
            self.as_connection.list_extend(key, "contact_no", [999], {}, policy, "")

        assert "list_extend() takes at most 5 arguments (6 given)" in str(typeError.value)

    def test_neg_list_extend_policy_is_string(self):
        """
        Invoke list_extend() with policy is string
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_extend(key, "contact_no", [85], {}, "")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "policy must be a dict"

    def test_neg_list_extend_key_is_none(self):
        """
        Invoke list_extend() with key is none
        """
        try:
            self.as_connection.list_extend(None, "contact_no", [45])

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "key is invalid"

    def test_neg_list_extend_bin_is_none(self):
        """
        Invoke list_extend() with bin is none
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_extend(key, None, ["str"])

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Bin name should be of type string"

    def test_neg_list_extend_with_string_instead_of_list(self):
        """
        Invoke list_extend() with string is passed in place of list
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_extend(key, "contact_no", "str")

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Items should be of type list"

    def test_neg_list_extend_meta_type_integer(self):
        """
        Invoke list_extend() with metadata input is of type integer
        """
        key = ("test", "demo", 1)
        try:
            self.as_connection.list_extend(key, "contact_no", [85], 888)

        except e.ParamError as exception:
            assert exception.code == -2
            assert exception.msg == "Metadata should be of type dictionary"



import socket
import time
import os
import sys
from distutils.version import LooseVersion

import pytest
from _pytest.terminal import TerminalReporter  # noqa: F401
from collections import namedtuple
from itertools import groupby
import tracemalloc

from . import invalid_data
from .test_base_class import TestBaseClass

aerospike = pytest.importorskip("aerospike")

test_memleak = int(os.environ.get("TEST_MEMLEAK", 0))

if test_memleak == 1:
    from psutil import Process

    _proc = Process(os.getpid())

    LEAK_LIMIT = 0

    def get_consumed_ram():
        return _proc.memory_info().rss

    START = "START"
    END = "END"
    ConsumedRamLogEntry = namedtuple("ConsumedRamLogEntry", ("nodeid", "on", "consumed_ram"))
    consumed_ram_log = []
    ConsumedTracemallocLogEntry = namedtuple("ConsumedTracemallocLogEntry", ("nodeid", "on", "consumed_tracemalloc"))
    consumed_tracemalloc_log = []

    tracemalloc.start(10)
    snapshot1 = []
    snapshot2 = []

    @pytest.hookimpl(hookwrapper=True)
    def pytest_terminal_summary(terminalreporter):  # type: (TerminalReporter) -> generator # noqa: F821
        yield

        # you can do here anything - I just print report info
        print("*" * 8 + "HERE CUSTOM LOGIC" + "*" * 8)

        for failed in terminalreporter.stats.get("failed", []):  # type: TestReport # noqa: F821
            print("failed! node_id:%s, duration: %s" % (failed.nodeid, failed.duration))

        for passed in terminalreporter.stats.get("passed", []):  # type: TestReport # noqa: F821
            print(
                "passed! node_id:%s, duration: %s, details: %s" % (passed.nodeid, passed.duration, str(passed.longrepr))
            )

        grouped = groupby(consumed_ram_log, lambda entry: entry.nodeid)
        for nodeid, (start_entry, end_entry) in grouped:
            leaked = end_entry.consumed_ram - start_entry.consumed_ram
            if leaked > LEAK_LIMIT:
                terminalreporter.write("LEAKED {}KB in {}\n".format(leaked / 1024, nodeid))

        tmgrouped = groupby(consumed_tracemalloc_log, lambda entry: entry.nodeid)
        for nodeid, (start_entry, end_entry) in tmgrouped:
            stats = end_entry.consumed_tracemalloc.compare_to(start_entry.consumed_tracemalloc, "lineno")
            print(f"{nodeid}:")
            for stat in stats[:3]:
                print(stat)
                # terminalreporter.write(stats)

    def pytest_runtest_setup(item):

        log_entry = ConsumedRamLogEntry(item.nodeid, START, get_consumed_ram())
        consumed_ram_log.append(log_entry)

        tmlog_entry = ConsumedTracemallocLogEntry(item.nodeid, START, tracemalloc.take_snapshot())
        consumed_tracemalloc_log.append(tmlog_entry)

    def pytest_runtest_teardown(item):

        log_entry = ConsumedRamLogEntry(item.nodeid, END, get_consumed_ram())
        consumed_ram_log.append(log_entry)

        tmlog_entry = ConsumedTracemallocLogEntry(item.nodeid, END, tracemalloc.take_snapshot())
        consumed_tracemalloc_log.append(tmlog_entry)


def compare_server_versions(version1, version2):
    """
    Compare two strings version1 and version 2

    Returns:
    -1 if version1 < version2
    0 if version1 == version2
    1 if version1 > version2
    """
    version1_pre = "pre" in version1
    version2_pre = "pre" in version2

    # Remove any suffix and build version of that
    loose_version1 = LooseVersion(version1.split("-")[0])
    loose_version2 = LooseVersion(version2.split("-")[0])

    if loose_version1 < loose_version2:
        return -1
    elif loose_version1 > loose_version2:
        return 1

    # Versions without suffix match

    # Both are preleases or neither is a prelease
    if version1_pre == version2_pre:
        return 0

    # Version 1 is pre release
    if version1_pre:
        return -1

    return 1


def wait_for_port(address, port, interval=0.1, timeout=60):
    """Wait for a TCP / IP port to accept a connection.

    : param port: The port to check.
    : param interval: The interval(seconds) between checks.
    : param timeout: The total time(seconds) to check before quiting.
    """
    start = time.time()
    while True:
        current = time.time()
        if current - start >= timeout:
            break
        try:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.connect((address, port))
            s.close()
            return True
        except Exception:
            pass
        time.sleep(interval)
    return False


@pytest.fixture(scope="class")
def as_connection(request):
    config = TestBaseClass.get_connection_config()
    lua_user_path = os.path.join(sys.exec_prefix, "aerospike", "usr-lua")
    lua_info = {"user_path": lua_user_path}
    config["lua"] = lua_info
    # print(config)
    as_client = None
    if len(config["hosts"]) == 2:
        for (a, p) in config["hosts"]:
            wait_for_port(a, p)
    # We are using tls otherwise, so rely on the server being ready

    if config["user"] is None and config["password"] is None:
        as_client = aerospike.client(config).connect()
    else:
        as_client = aerospike.client(config).connect(config["user"], config["password"])

    request.cls.skip_old_server = True
    request.cls.server_version = []
    versioninfo = as_client.info_all("build")
    for keys in versioninfo:
        for value in versioninfo[keys]:
            if value is not None:
                version_str = value.strip()
                versionlist = version_str.split(".")
                request.cls.string_server_version = version_str
                request.cls.server_version = [int(n) for n in versionlist[:2]]
                if (int(versionlist[0]) > 3) or (int(versionlist[0]) == 3 and int(versionlist[1]) >= 7):
                    request.cls.skip_old_server = False
                TestBaseClass.major_ver = int(versionlist[0])
                TestBaseClass.minor_ver = int(versionlist[1])

    request.cls.as_connection = as_client

    def close_connection():
        as_client.close()

    request.addfinalizer(close_connection)
    return as_client


@pytest.fixture(scope="class")
def connection_with_udf(request, as_connection):
    """
    Injects an as_client() as a dependency, loads a udf to it,
    and at the end of the test class removes it.
    Note: if the requesting class does not have a class attr:
    `udf_to_load`, this is essentially a noop
    """
    udf_status = {"loaded": False, "name": None}
    # if the class doesn't have the correct information,
    # don't bother loading a UDF
    if hasattr(request.cls, "udf_to_load"):
        udf_status["name"] = request.cls.udf_to_load
        as_connection.udf_put(udf_status["name"], 0, {})
        udf_status["loaded"] = True

    # Yield to the requesting context
    yield as_connection

    # If a UDF has been loaded, remove it
    if udf_status["loaded"]:
        try:
            as_connection.udf_remove(udf_status["name"])
        except Exception:  # If this fails, it has already been removed
            pass


@pytest.fixture(scope="class")
def connection_with_config_funcs(request, as_connection):
    """
    Injects a connected as_client() as a dependency, and runs arbitrary
    setup functions on it, ideally these would be expensive functions
    which should only be run once per test category: indexing
    loading udfs. Then yields the connection back to the requesting context
    """
    setup_status = False
    if hasattr(request.cls, "connection_setup_functions"):
        for func in request.cls.connection_setup_functions:
            func(as_connection)
        setup_status = True

    # Yield to the requesting context
    yield as_connection

    # If any setup was done, run the corresponding teardown functions
    if setup_status and hasattr(request.cls, "connection_teardown_functions"):
        for func in request.cls.connection_teardown_functions:
            func(as_connection)
        setup_status = False


@pytest.fixture()
def put_data(request):
    put_data.key = None
    put_data.keys = []
    put_data.client = None

    def put_key(client, _key, _record, _meta=None, _policy=None):
        put_data.key = _key
        put_data.client = client
        try:
            client.remove(_key)
        except Exception:
            pass
        put_data.keys.append(put_data.key)
        return client.put(_key, _record, _meta, _policy)

    def remove_key():
        try:
            # pytest.set_trace()
            for key in put_data.keys:
                put_data.client.remove(key)
        except Exception:
            pass

    request.addfinalizer(remove_key)
    return put_key


@pytest.fixture(scope="class")
def connection_config(request):
    """
    Sets the class attribute to be the config object passed in
     to create the as_connection
    """
    config = TestBaseClass.get_connection_config()
    request.cls.connection_config = config


@pytest.fixture(params=invalid_data.INVALID_KEYS)
def invalid_key(request):
    yield request.param


# aerospike.set_log_level(aerospike.LOG_LEVEL_DEBUG)
# aerospike.set_log_handler(None)



# -*- coding: utf-8 -*-


import pytest
from aerospike_helpers.operations import map_operations as mh

import aerospike

from aerospike_helpers.expressions import Add, Cond, Def, GE, IntBin, LT, Let, Mul, Unknown, Var
from aerospike_helpers.operations import expression_operations as expressions
from aerospike import exception as e
from .test_base_class import TestBaseClass


class TestBatchExpressionsOperations(TestBaseClass):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):
        # TODO this should be changed to 6.0 before release.
        if self.server_version < [5, 6]:
            pytest.mark.xfail(reason="Servers older than 5.6 do not support batch get ops.")
            pytest.xfail()

        self.test_ns = "test"
        self.test_set = "demo"
        self.keys = []
        self.rec_count = 5

        for i in range(self.rec_count):
            key = ("test", "demo", i)
            rec = {
                "name": "name10",
                "t": True,
                "age": 10,
                "balance": 100,
                "key": 10,
                "ilist_bin": [
                    1,
                    2,
                    6,
                ],
                "imap_bin": {
                    1: 1,
                    2: 2,
                    3: 6,
                },
            }
            as_connection.put(key, rec)
            self.keys.append(key)

        def teardown():
            for i in range(self.rec_count):
                key = ("test", "demo", i)
                as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "expr, flags, name, expected",
        [
            (
                Let(Def("bal", IntBin("balance")), Var("bal")),
                aerospike.EXP_READ_DEFAULT,
                "test_name",
                {"test_name": 100},
            ),
            (
                Let(Def("bal", IntBin("balance")), Cond(GE(Var("bal"), 50), Add(Var("bal"), 50), Unknown())),
                aerospike.EXP_READ_DEFAULT,
                "test_name2",
                {"test_name2": 150},
            ),
            (
                Let(
                    Def("bal", IntBin("balance")),
                    Def("age", IntBin("age")),
                    Cond(GE(Var("bal"), 50), Mul(Var("bal"), Var("age")), Unknown()),
                ),
                aerospike.EXP_READ_DEFAULT,
                "test_mul1",
                {"test_mul1": 1000},
            ),
        ],
    )
    def test_read_pos(self, expr, flags, name, expected):
        """
        Test expression read operation with correct parameters.
        """
        ops = [expressions.expression_read(name, expr.compile(), flags)]
        meta = {"gen": 1}

        res = self.as_connection.batch_get_ops(self.keys, ops, meta)
        """
        res are in the format of (status-tuple, ((meta-dict, result-dict), status-tuple, exception), ...)
        """
        # print(res)
        for i in range(self.rec_count):
            assert res[0][2] == expected

    @pytest.mark.parametrize(
        "expr, flags, name, expected",
        [
            ("bad_expr", aerospike.EXP_READ_DEFAULT, "test_name2", e.ParamError),
            (
                Let(Def("bal", IntBin("balance")), Cond(LT(Var("bal"), 50), Add(Var("bal"), 50), Unknown())).compile(),
                "bad_flags",
                "test_name3",
                e.ParamError,
            ),
        ],
    )
    def test_read_neg(self, expr, flags, name, expected):
        """
        Test expression read operation expecting failure.
        """
        with pytest.raises(expected):
            ops = [expressions.expression_read(name, expr, flags)]
            self.as_connection.batch_get_ops(self.keys, ops)
            # print("test_read_neg: ", res)

    def test_batch_result_output_format(self):
        # pp = pprint.PrettyPrinter(2, 80)
        policy = {"key": aerospike.POLICY_KEY_SEND}
        map_policy = {
            "map_write_mode": aerospike.MAP_UPDATE,
            "map_order": aerospike.MAP_KEY_ORDERED,
        }

        key1 = ("test", "demo", "batch-ops-k1")
        scores1 = {"u1": 123, "u2": 234, "u7": 789, "u8": 890, "u9": 901}
        ops = [mh.map_put_items("scores", scores1, map_policy)]
        self.as_connection.operate(key1, ops, policy=policy)
        key2 = ("test", "demo", "batch-ops-k2")
        scores2 = {"z1": 321, "z2": 432, "z7": 987, "z8": 98, "z9": 109}
        ops = [mh.map_put_items("scores", scores2, map_policy)]
        self.as_connection.operate(key2, ops, policy=policy)

        ops = [mh.map_get_by_rank_range("scores", -3, 3, aerospike.MAP_RETURN_KEY_VALUE)]
        non_existent_key = ("test", "demo", "batch-ops-non_existent_key")
        rec = self.as_connection.batch_get_ops([key1, key2, non_existent_key], ops, policy)

        # print("\nThe record from batch_get_ops")
        # pp.pprint(rec)

        assert rec[0][-1] is not None
        assert rec[1][-1] is not None
        assert rec[2][-2] == e.RecordNotFound
        assert rec[2][-1] is None

        # rec = self.as_connection.select_many([non_existent_key], ['name'])
        # print("\nFor comparison, here's batch-read (select_many) is an array of records")
        # pp.pprint(rec)

        # rec = self.as_connection.get_many([key1, key2, non_existent_key], policy)
        # print("\nFor comparison, here's batch-read (get_many) is an array of records")
        # pp.pprint(rec)

        # print("\nThe record coming from opreate")
        # rec = self.as_connection.operate(key1, ops, policy=policy)
        # pp.pprint(rec)

        self.as_connection.remove(key1)
        self.as_connection.remove(key2)



# -*- coding: utf-8 -*-
import pytest
from aerospike import exception as e

import aerospike


class TestMapBasics(object):
    @pytest.fixture(autouse=True)
    def setup(self, request, as_connection):

        key = ("test", "map_test", 1)
        rec = {"blank": "blank"}
        as_connection.put(key, rec)

        def teardown():
            """
            Teardown method.
            """
            key = ("test", "map_test", 1)
            self.as_connection.remove(key)

        request.addfinalizer(teardown)

    @pytest.mark.parametrize(
        "key, binname, map_key, map_value, expected",
        [
            (("test", "map_test", 1), "my_map", "age", 97, [("age", 97)]),
            (
                ("test", "map_test", 1),
                "my_map",
                "bytes",
                bytearray("1234", "utf-8"),
                [("bytes", bytearray("1234", "utf-8"))],
            ),
            (("test", "map_test", 1), "my_map", bytearray("1234", "utf-8"), 1234, [(bytearray("1234", "utf-8"), 1234)]),
            (("test", "map_test", 1), "my_map2", "name", "dean", [("name", "dean")]),
        ],
    )
    def test_pos_map_put_get(self, key, binname, map_key, map_value, expected):
        """
        Invoke map_put() insert value with correct parameters
        """
        self.as_connection.map_put(key, binname, map_key, map_value)
        bins = self.as_connection.map_get_by_key(key, binname, map_key, aerospike.MAP_RETURN_KEY_VALUE)
        self.as_connection.map_clear(key, binname)
        assert bins == expected

    @pytest.mark.parametrize(
        "key, binname, map_items, map_key, expected",
        [
            (("test", "map_test", 1), "my_map", {"item1": "value1", "item2": "value2"}, "item2", [("item2", "value2")]),
            (("test", "map_test", 1), "my_map", {15: "value15", 23: "value23"}, 15, [(15, "value15")]),
            (
                ("test", "map_test", 1),
                "my_map",
                {15: bytearray("1234", "utf-8"), 23: "value23"},
                15,
                [(15, bytearray("1234", "utf-8"))],
            ),
        ],
    )
    def test_pos_map_put_items(self, key, binname, map_items, map_key, expected):
        self.as_connection.map_put_items(key, binname, map_items)
        bins = self.as_connection.map_get_by_key(key, binname, map_key, aerospike.MAP_RETURN_KEY_VALUE)
        self.as_connection.map_clear(key, binname)
        assert bins == expected

    def test_pos_map_clear(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, "age", 97)
        size = self.as_connection.map_size(key, binname)
        self.as_connection.map_clear(key, binname)
        newSize = self.as_connection.map_size(key, binname)
        assert size == 1
        assert newSize == 0

    @pytest.mark.parametrize(
        "key, binname, map_key, map_value, map_increment, expected",
        [
            (("test", "map_test", 1), "my_map", "int", 99, 1, 100),
            (("test", "map_test", 1), "my_map", bytearray("1234", "utf-8"), 99, 1, 100),
            (("test", "map_test", 1), "my_map2", "float", 2.1, 1.2, 3.3),
        ],
    )
    def test_pos_map_increment(self, key, binname, map_key, map_value, map_increment, expected):
        self.as_connection.map_put(key, binname, map_key, map_value)
        self.as_connection.map_increment(key, binname, map_key, map_increment)
        value = self.as_connection.map_get_by_key(key, binname, map_key, aerospike.MAP_RETURN_VALUE)
        self.as_connection.map_clear(key, binname)
        assert value == expected

    @pytest.mark.parametrize(
        "key, binname, map_key, map_value, map_decrement, expected",
        [
            (("test", "map_test", 1), "my_map", "int", 99, 1, 98),
            (("test", "map_test", 1), "my_map2", "float", 2.1, 1.1, 1.0),
            (("test", "map_test", 1), "my_map", bytearray("1234", "utf-8"), 100, 1, 99),
        ],
    )
    def test_pos_map_decrement(self, key, binname, map_key, map_value, map_decrement, expected):
        self.as_connection.map_put(key, binname, map_key, map_value)
        self.as_connection.map_decrement(key, binname, map_key, map_decrement)
        value = self.as_connection.map_get_by_key(key, binname, map_key, aerospike.MAP_RETURN_VALUE)
        self.as_connection.map_clear(key, binname)
        assert value == expected

    def test_pos_map_remove_by_key(self):
        key = ("test", "map_test", 1)
        binname = "my_map"

        self.as_connection.map_put(key, binname, "age", 97)
        value = self.as_connection.map_remove_by_key(key, binname, "age", aerospike.MAP_RETURN_VALUE)
        value2 = self.as_connection.map_remove_by_key(key, binname, "age", aerospike.MAP_RETURN_VALUE)
        assert value == 97
        assert value2 is None

        byte_map_bin = bytearray("1234", "utf-8")
        self.as_connection.map_put(key, binname, byte_map_bin, 97)
        value = self.as_connection.map_remove_by_key(key, binname, byte_map_bin, aerospike.MAP_RETURN_VALUE)
        value2 = self.as_connection.map_remove_by_key(key, binname, byte_map_bin, aerospike.MAP_RETURN_VALUE)
        assert value == 97
        assert value2 is None

    def test_pos_map_remove_by_key_list(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        self.as_connection.map_remove_by_key_list(key, binname, ["item1", "item2"], aerospike.MAP_RETURN_NONE)
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_key_list_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, bytearray("1234", "utf-8"), 5)
        self.as_connection.map_put(key, binname, bytearray("1235", "utf-8"), 6)
        self.as_connection.map_remove_by_key_list(
            key, binname, [bytearray("1234", "utf-8"), bytearray("1235", "utf-8")], aerospike.MAP_RETURN_NONE
        )
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_key_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        self.as_connection.map_remove_by_key_range(key, binname, "i", "j", aerospike.MAP_RETURN_NONE)
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_key_range_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, bytearray([1, 2, 3, 4]), 5)
        self.as_connection.map_put(key, binname, bytearray([1, 9, 9, 9]), 6)

        self.as_connection.map_remove_by_key_range(
            key, binname, bytearray([1, 2, 3, 4]), bytearray([2, 0, 0, 0]), aerospike.MAP_RETURN_NONE
        )

        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_value(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        count1 = self.as_connection.map_remove_by_value(key, binname, "value1", aerospike.MAP_RETURN_COUNT)
        count2 = self.as_connection.map_remove_by_value(key, binname, "value2", aerospike.MAP_RETURN_COUNT)
        size = self.as_connection.map_size(key, binname)
        assert count1 == 1
        assert count2 == 1
        assert size == 0

    def test_pos_map_remove_by_value_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        val1 = bytearray([1, 2, 3, 4])
        val2 = bytearray([1, 2, 3, 5])
        self.as_connection.map_put_items(key, binname, {"item1": val1, "item2": val2})
        count1 = self.as_connection.map_remove_by_value(key, binname, val1, aerospike.MAP_RETURN_COUNT)
        count2 = self.as_connection.map_remove_by_value(key, binname, val2, aerospike.MAP_RETURN_COUNT)
        size = self.as_connection.map_size(key, binname)
        assert count1 == 1
        assert count2 == 1
        assert size == 0

    def test_pos_map_remove_by_value_list(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        self.as_connection.map_remove_by_value_list(key, binname, ["value1", "value2"], aerospike.MAP_RETURN_NONE)
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_value_list_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        val1 = bytearray([1, 2, 3, 4])
        val2 = bytearray([1, 2, 3, 5])
        self.as_connection.map_put_items(key, binname, {"item1": val1, "item2": val2})
        self.as_connection.map_remove_by_value_list(key, binname, [val1, val2], aerospike.MAP_RETURN_NONE)
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_value_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        self.as_connection.map_remove_by_value_range(key, binname, "v", "w", aerospike.MAP_RETURN_NONE)
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_value_range_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        val1 = bytearray([1, 2, 3, 4])
        val2 = bytearray([1, 2, 3, 6])
        self.as_connection.map_put_items(key, binname, {"item1": val1, "item2": val2})
        self.as_connection.map_remove_by_value_range(
            key, binname, bytearray([1, 2, 3, 3]), bytearray([1, 2, 3, 7]), aerospike.MAP_RETURN_NONE
        )
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_remove_by_index(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_set_policy(key, binname, {"map_sort": aerospike.MAP_KEY_ORDERED})
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        value1 = self.as_connection.map_remove_by_index(key, binname, 1, aerospike.MAP_RETURN_VALUE)
        value2 = self.as_connection.map_remove_by_index(key, binname, 0, aerospike.MAP_RETURN_VALUE)
        size = self.as_connection.map_size(key, binname)
        assert size == 0
        assert value1 == "value2"
        assert value2 == "value1"

    def test_pos_map_remove_by_index_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        count = self.as_connection.map_remove_by_index_range(key, binname, 0, 2, aerospike.MAP_RETURN_COUNT)
        size = self.as_connection.map_size(key, binname)
        assert count == 2
        assert size == 0

    def test_pos_map_remove_by_rank(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": 10, "item2": 11})
        theKey = self.as_connection.map_remove_by_rank(key, binname, 1, aerospike.MAP_RETURN_KEY)
        size = self.as_connection.map_size(key, binname)
        self.as_connection.map_clear(key, binname)
        assert theKey == "item2"
        assert size == 1

    def test_pos_map_remove_by_rank_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": 10, "item2": 11, "item3": 12})
        count = self.as_connection.map_remove_by_rank_range(key, binname, 0, 2, aerospike.MAP_RETURN_COUNT)
        size = self.as_connection.map_size(key, binname)
        self.as_connection.map_clear(key, binname)
        assert count == 2
        assert size == 1

    def test_pos_map_get_by_key_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        values = self.as_connection.map_get_by_key_range(key, binname, "i", "j", aerospike.MAP_RETURN_VALUE)
        self.as_connection.map_clear(key, binname)
        assert values == ["value1", "value2"]

    def test_pos_map_get_by_key_range_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, bytearray([1, 2, 3, 4]), 5)
        self.as_connection.map_put(key, binname, bytearray([1, 9, 9, 9]), 6)

        values = self.as_connection.map_get_by_key_range(
            key, binname, bytearray([1, 2, 3, 4]), bytearray([2, 0, 0, 0]), aerospike.MAP_RETURN_VALUE
        )
        self.as_connection.map_clear(key, binname)
        assert values == [5, 6]

    def test_pos_map_get_by_value(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_set_policy(key, binname, {"map_sort": aerospike.MAP_KEY_ORDERED})
        self.as_connection.map_put_items(key, binname, {"item1": "value2", "item2": "value2"})
        key1 = self.as_connection.map_get_by_value(key, binname, "value2", aerospike.MAP_RETURN_KEY)
        self.as_connection.map_clear(key, binname)
        assert key1 == ["item1", "item2"]

    def test_pos_map_get_by_value_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2", "item3": "value3"})
        keys = self.as_connection.map_get_by_value_range(key, binname, "value1", "value3", aerospike.MAP_RETURN_KEY)
        self.as_connection.map_clear(key, binname)
        assert keys == ["item1", "item2"]

    def test_pos_map_get_by_value_range_bytes(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, 5, bytearray([1, 2, 3, 4]))
        self.as_connection.map_put(key, binname, 6, bytearray([1, 9, 9, 9]))

        values = self.as_connection.map_get_by_value_range(
            key, binname, bytearray([1, 2, 3, 4]), bytearray([2, 0, 0, 0]), aerospike.MAP_RETURN_KEY
        )
        self.as_connection.map_clear(key, binname)
        assert values == [5, 6]

    @pytest.mark.parametrize("policy, meta", [(None, None), ({"total_timeout": 180000}, {"ttl": -1})])
    def test_map_get_by_value_list(self, policy, meta):
        key = ("test", "map_test", 1)
        binname = "my_map"
        map_entry = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 1}
        self.as_connection.put(key, {binname: map_entry})
        if policy:
            fetched_keys = self.as_connection.map_get_by_value_list(
                key, binname, [1, 3, 4], aerospike.MAP_RETURN_KEY, policy=policy, meta=meta
            )
        else:
            fetched_keys = self.as_connection.map_get_by_value_list(key, binname, [1, 3, 4], aerospike.MAP_RETURN_KEY)

        assert set(fetched_keys) == set(["a", "c", "d", "e"])
        self.as_connection.map_clear(key, binname)

    def test_map_get_by_value_list_no_matching_items(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        map_entry = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 1}
        self.as_connection.put(key, {binname: map_entry})
        fetched_keys = self.as_connection.map_get_by_value_list(
            key, binname, [7, 8, 9], aerospike.MAP_RETURN_KEY, policy={"total_timeout": 180000}, meta={"ttl": -1}
        )

        assert fetched_keys == []
        self.as_connection.map_clear(key, binname)

    @pytest.mark.parametrize("policy, meta", [(None, None), ({"total_timeout": 180000}, {"ttl": -1})])
    def test_map_get_by_key_list(self, policy, meta):
        key = ("test", "map_test", 1)
        binname = "my_map"
        map_entry = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 1}
        self.as_connection.put(key, {binname: map_entry})
        if policy:
            fetched_vals = self.as_connection.map_get_by_key_list(
                key, binname, ["a", "c", "e"], aerospike.MAP_RETURN_VALUE, policy=policy, meta=meta
            )
        else:
            fetched_vals = self.as_connection.map_get_by_key_list(
                key, binname, ["a", "c", "e"], aerospike.MAP_RETURN_VALUE
            )

        fetched_vals.sort()
        assert fetched_vals == [1, 1, 3]

        self.as_connection.map_clear(key, binname)

    def test_map_get_by_key_list_no_matching(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        map_entry = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 1}
        self.as_connection.put(key, {binname: map_entry})
        fetched_vals = self.as_connection.map_get_by_key_list(
            key, binname, ["aero", "spike"], aerospike.MAP_RETURN_VALUE
        )

        assert fetched_vals == []

        self.as_connection.map_clear(key, binname)

    def test_pos_map_get_by_index(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_set_policy(key, binname, {"map_sort": aerospike.MAP_KEY_ORDERED})
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        value1 = self.as_connection.map_get_by_index(key, binname, 0, aerospike.MAP_RETURN_VALUE)
        value2 = self.as_connection.map_get_by_index(key, binname, 1, aerospike.MAP_RETURN_VALUE)
        self.as_connection.map_clear(key, binname)
        assert value1 == "value1"
        assert value2 == "value2"

    def test_pos_map_get_get_by_index_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_set_policy(key, binname, {"map_sort": aerospike.MAP_KEY_ORDERED})
        self.as_connection.map_put_items(key, binname, {"item1": "value1", "item2": "value2"})
        values = self.as_connection.map_get_by_index_range(key, binname, 0, 2, aerospike.MAP_RETURN_VALUE)
        self.as_connection.map_clear(key, binname)
        assert values == ["value1", "value2"]

    def test_pos_map_get_by_rank(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": 10, "item2": 11})
        theKey = self.as_connection.map_get_by_rank(key, binname, 1, aerospike.MAP_RETURN_KEY)
        self.as_connection.map_clear(key, binname)
        assert theKey == "item2"

    def test_pos_map_get_by_rank_range(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": 10, "item2": 11, "item3": 12})
        count = self.as_connection.map_get_by_rank_range(key, binname, 0, 2, aerospike.MAP_RETURN_COUNT)
        self.as_connection.map_clear(key, binname)
        assert count == 2

    def test_pos_map_insert_empty_map_into_cleared_map(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put_items(key, binname, {"item1": 3})
        self.as_connection.map_clear(key, binname)
        self.as_connection.map_put_items(key, binname, {})
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_pos_map_insert_none_value(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, "my_key", None)
        size = self.as_connection.map_size(key, binname)
        assert size == 1
        value = self.as_connection.map_get_by_key(key, binname, "my_key", aerospike.MAP_RETURN_VALUE)
        assert value is None
        self.as_connection.map_clear(key, binname)

    def test_pos_map_insert_empty_value(self):
        key = ("test", "map_test", 1)
        binname = "my_map"
        self.as_connection.map_put(key, binname, "my_key", {})
        size = self.as_connection.map_size(key, binname)
        assert size == 1
        value = self.as_connection.map_get_by_key(key, binname, "my_key", aerospike.MAP_RETURN_VALUE)
        assert value == {}
        self.as_connection.map_clear(key, binname)

    def test_map_get_by_value_list_non_list(self):
        key = ("test", "demo", 1)
        binname = "my_map"
        with pytest.raises(e.ParamError):
            self.as_connection.map_get_by_value_list(key, binname, "A value", aerospike.MAP_RETURN_KEY)

    def test_map_get_by_key_list_non_list(self):
        key = ("test", "demo", 1)
        binname = "my_map"
        with pytest.raises(e.ParamError):
            self.as_connection.map_get_by_key_list(key, binname, "A key, and another", aerospike.MAP_RETURN_KEY)

    def test_map_get_by_value_list_no_return_type(self):
        key = ("test", "demo", 1)
        binname = "my_map"
        with pytest.raises(TypeError):
            self.as_connection.map_get_by_value_list(key, binname, [1, 2, 3])

    def test_map_get_by_key_list_no_return_type(self):
        key = ("test", "demo", 1)
        binname = "my_map"
        with pytest.raises(TypeError):
            self.as_connection.map_get_by_key_list(key, binname, ["k1", "k2", "k3"])

    def test_map_size_nonexistent_bin(self):
        key = ("test", "map_test", 1)
        binname = "non_a_real_bin"
        size = self.as_connection.map_size(key, binname)
        assert size == 0

    def test_map_insert_invalid_key(self):
        key = ()
        binname = "my_map"
        with pytest.raises(e.ParamError):
            self.as_connection.map_put(key, binname, "my_key", {})

    @pytest.mark.parametrize(
        "method_name",
        [
            "map_set_policy",
            "map_put",
            "map_put_items",
            "map_increment",
            "map_decrement",
            "map_size",
            "map_clear",
            "map_remove_by_key",
            "map_remove_by_key_list",
            "map_remove_by_key_range",
            "map_remove_by_value",
            "map_remove_by_value_list",
            "map_remove_by_value_range",
            "map_remove_by_index",
            "map_remove_by_rank",
            "map_remove_by_rank_range",
            "map_get_by_key",
            "map_get_by_key_range",
            "map_get_by_value",
            "map_get_by_value_range",
            "map_get_by_index",
            "map_get_by_index_range",
            "map_get_by_rank",
            "map_get_by_rank_range",
            "map_get_by_key_list",
            "map_get_by_value_list",
        ],
    )
    def test_call_to_methods_with_no_args(self, method_name):
        """
        Call each of the map_* methods with no arguments
        """
        method = getattr(self.as_connection, method_name)
        with pytest.raises(TypeError):
            method()
